//#include "stdafx.h" // Other headers defined in stdafx.h

#include <conio.h>
#include <iostream>
#include <ctime>

#include "uCalc.h"
#include "UserCallbacks.h"

void Setup() {
	// The MyErrorHandler() function will be called whenever and wherever an error (such as "Syntax Error") is raised.	
	uc.AddErrorHandler(MyErrorHandler);			

	// SetOutput lets you configure the formatting of results returned with EvalStr() or EvaluateStr()			
	uc.SetOutput(MyNumericFormat, "Double");

	// MyArea() returns the product of multiplying the two arguments.
	// MyLeft() returns the left-most characters of a string.  The second argument represents the number of characters.
	// MsgBox() displays a message box.  The 1st argument is required.  The others are optional.
	uc.DefineFunction("MyArea(Length, Width)", MyArea);
	uc.DefineFunction("MyLeft(Text As String, Count As Int32) As String", MyLeft);
	uc.DefineFunction("MyMsgBox(Prompt As String, Buttons As Int32 = 0, Title As String = 'uCalc')", MyMsgBox);
	uc.ParserTransform().FromTo("MsgBox ::= MyMsgBox");

	// The three consecutive dots "..." mean that MyAverage can take any number of arguments
	// (however, because "x" is specified, there must be at least one argument).
	uc.DefineFunction("MyAverage(x ...)", MyAverage);

	// Two versions of MyIIF can co-exist peacefully with the same name because
	// they are defined with different argument types (numeric vs string).
	// By passing the last two arguments ByExpr, the callback can chose to evaluate only one of of them.  
	uc.DefineFunction("MyIIf(cond, ByExpr TruePart, ByExpr FalsePart)", MyIIF_Numeric);
	uc.DefineFunction("MyIIf(cond, ByExpr TruePart As String, ByExpr FalsePart As String) As String", MyIIF_String);

	// The following defines the "*" operator so that MyString * n returns MyString repeated n times.
	// For instance, "He " * 3, would return "He He He".  20 represents the precedence level.  It is
	// arbitrarily set at the same level as that of the numeric "*" multiplication operator.
	uc.DefineOperator("{MyString As String} * {Number} As String", uCalc::Item(uc.HandleOf("*")).Precedence(), GroupLeftToRight, StringRepeat);

	// This definition is for a summation.  See the callback ucSum() routine for the
	// actual code, which runs a loop that adds up the total for the expression in
	// the first argument a number of times based on the second and third arguments.
	// For instance Sum(g^2+1, 1, 5, 1, g) returns 60 and Sum(x^2, 1, 10) returns 385.
	// The last two arguments are optional, so they default to 1 and x if omitted.
	//
	// The first argument is passed "ByExpr".  So instead of being evaluated before
	// being passed to the callback the way an ordinary argument would be, a handle
	// for the expression is passed so that the callback can evaluate it (in this
	// case numerous times).
	//
	// The last argument is passed "ByHandle".  This causes the callback to receive
	// a handle for the variable being passed, in such a way that it can be linked to
	// the summation counter, and integrated into the expression in the first argument.
	//
	// The actual function being defined is Sum_().  Then a syntax construct named
	// Sum() is defined in such a way that the last argument gets defined as a local
	// variable.  So if you evaluate Sum(x^2, 5, 10, 1, x), the local "x" in this
	// expression will not interfere with a pre-existing variable named x.  Also you
	// do not need to declare a variable ahead of time to use it as a counter for Sum().
	uc.DefineFunction("Sum_(ByExpr Expr, Start, Finish, Step, ByHandle Var)", ucSum);
	uc.ParserTransform().FromTo("Sum({Expr}, {Start}, {Finish} [, {Step=1} [, {Var=x}]])"
		"::= Local({Var}, Sum_({Expr}, {Start}, {Finish}, {Step}, {Var}))");

	// The following routine solves an equation.
	// The callback code is based on the Bisection Method algorithm.
	// The concept here is very similar to that of Sum().
	// Two syntax constructs are defined.  The second one rearranges the equation
	// if it includes an equal sign.
	// For instance Solve(x^2 = 9+x) becomes Solve(x^2 - (9+x))
	// Solve(x^2 + 1 = 26) returns 5.
	// Solve(x^2 + 1 = 26, -1000, 0) returns -5.
	uc.DefineFunction("Solve_(ByExpr Expr, a, b, ByHandle Var)", ucSolve);
	uc.ParserTransform().FromTo("Solve({Expr} [, {a=-100000000} [, {b=100000000} [, {Var=x}]]]) "
		"::= Local({Var}, Solve_({Expr}, {a}, {b}, {Var}))");
	uc.ParserTransform().FromTo("Solve({Left} = {Right} [, {etc}]) ::= Solve({Left}-({Right}) {etc: , {etc}})");
}

void ShowExpressionExamples() {
	system("CLS");
	cout << ".5 + 10" << endl;
	cout << "5*(((10+3)!-1) MOD 7)^2" << endl;
	cout << "MyArea(3,4)" << endl;
	cout << "MyArea(-3,4)" << endl;
	cout << "MyIIF(1 > 2, \"True\", \"False\")" << endl;
	cout << "Solve(x^2 + 1 = 26)" << endl;
	cout << "Sum(2*x + 5, 1, 10)" << endl;
	cout << "\"Hello \" + \"world\"" << endl;
	cout << "\"This\" > \"That\"" << endl;
	cout << "Asc(\"D\")" << endl;
	cout << "#hAE.B1 + #b1110.1001" << endl;
	cout << "MsgBox(\"Please send feedback at www.ucalc.com\")" << endl;
	cout << "MyAverage(1,2,3,4,5,6)" << endl;
	cout << "MyLeft(\"Hello There\", 7)" << endl;
	cout << "MsgBox(\"First box\") ; MsgBox(\"Second Box\",, \"Title\") ; 123" << endl;
	cout << "\"Hello \" * 3" << endl << endl << endl;
}

void ShowOutputFormatExamples() {
	system("CLS");
	cout << "The callback in this example uses sprintf_s() to format numeric results." << endl;
	cout << "Try examples such as:" << endl << endl;
	cout << "%f" << endl;
	cout << "%4.2f" << endl;
	cout << "Answer: %4.2f" << endl;
	cout << "%e" << endl;
	cout << "%a" << endl << endl << endl;
}

void ShowDefineExamples() {
	system("CLS");
	cout << "Var: MyString As String = \"Hello\"" << endl;
	cout << "Var: StringB = \"StringB is automatically defined as string because of quotes\"" << endl;
	cout << "Var: MyInt As Long = 123" << endl;
	cout << "Const: Pi = Atan(1) * 4" << endl;
	cout << "Func: f(x) = x^2" << endl;
	cout << "Func: g(x, y) = x * y" << endl;
	cout << "Func: Twice(x As Double) As Double = x * 2" << endl;
	cout << "Func: Twice(x As String) As String = x + x" << endl;
	cout << "Func: Area(x) = x ^ 2" << endl;
	cout << "Func: Area(x, y) = x * y" << endl;
	cout << "Func: Volume(r, h=0) = IIf(h==0, pi*r^3*4/3, pi*r^2*h/3)" << endl;
	cout << "Func: Factorial(x) = IIf(x>1, x*Factorial(x-1), 1)" << endl << endl << endl;
}

void Summation() {
	string UserExpression;
	double x, SumMax, SumTotal = 0;
	cout << "A summation will be performed." << endl;
	cout << "Enter an expression (Example: x^2+5*x-10): "; cin >> UserExpression;
	cout << "Go from 1 to how much (Example: 1000000)? "; cin >> SumMax;
	clock_t BeginTimer = clock();

	auto Variable = uc.DefineVariable("x", &x);
	auto Expression = uc.Parse(UserExpression);

	for(x = 1; x <= SumMax; x++)
	{
		//Variable.SetVariableValue(x);
		SumTotal = SumTotal + Expression.Evaluate();                
	}

	Expression.Release();
	Variable.Release();  			

	cout << "Total: " << SumTotal << endl;
	cout << "Elapsed time: " << double(clock()-BeginTimer) / CLOCKS_PER_SEC << endl << endl;
}

int main()
{	
	Setup();
	string Expression, Definition;	
	int Option;
      
	while (1) {

		cout << "Choose one of the following options: " << endl;
		cout << "1 Evaluate an expression" << endl;
		cout << "2 Change the output format" << endl;
		cout << "3 Define something (a function, variable, operator, etc)" << endl;
		cout << "4 Perform speed test (summation)" << endl;
		cout << "5 Quit" << endl;		

		switch(_getch()) {
		case '1':
			ShowExpressionExamples();
			cout << "Enter an expression, like one of the above: "; getline(cin, Expression);
			cout << "Result: " << uc.EvalStr(Expression) << endl;			
			break;
		case '2':
			ShowOutputFormatExamples();
			cout << "Enter a format for the numeric output: "; getline(cin, UserFormat);
			cout << endl << "Now try evaluating an expression." << endl;			
			break;
		case '3':
			ShowDefineExamples();
			cout << "Enter a variable, funciton, or operator defition: "; getline(cin, Definition);
			uc.Define(Definition);
			cout << endl << "Now try evaluating an expression with this definition." << endl;			
			break;
		case '4':
			Summation();
			break;			
		case '5':
			return 0;
			break;
		default:
			cout << "Please enter the numeric value corresponding to an option" << endl << endl;
			break;
		}

		cout << "Press a key to continue..."; _getch(); system("CLS");

	}


	return 0;
}

